@@ -6,7 +6,7 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.29';
 
 use File::Copy;
 use File::Path;
@@ -16,7 +16,7 @@ use LWP::Simple;
 use File::Basename;
 use File::Spec::Functions qw(rel2abs abs2rel);
 use Cwd;
-use File::Temp;
+use File::Temp qw(tempfile);
 
 our $DRY_RUN;
 our $CONFIRM;
@@ -77,11 +77,13 @@ cp rmf mkd cd make
 cdback download untar 
 pie slurp blurt mv tap 
 plough qquote quote perm_cp
+perm_get perm_set
 sysrun untar_in pick ask
 hammer say
 sudo_me bin_find
 fs_read_open fs_write_open pipe_copy
-snip
+snip password_read nice_time
+def_or blurt_atomic
 );
 
 our %EXPORTABLE = map { $_ => 1 } @EXPORTABLE;
@@ -194,8 +196,10 @@ sub cp {
 
     _confirm("cp $_[0] $_[1]") or return 1;
 
+    INFO "cp $_[0] $_[1]";
+
     File::Copy::copy @_ or 
-        get_logger("")->logcroak("Cannot copy $_[0] to $_[1] ($!)");
+        LOGCROAK("Cannot copy $_[0] to $_[1] ($!)");
 }
 
 =pod
@@ -214,8 +218,10 @@ sub mv {
 
     _confirm("mv $_[0] $_[1]") or return 1;
 
+    INFO "mv $_[0] $_[1]";
+
     File::Copy::move @_ or 
-        get_logger("")->logcroak("Cannot move $_[0] to $_[1] ($!)");
+        LOGCROAK("Cannot move $_[0] to $_[1] ($!)");
 }
 
 =pod
@@ -230,13 +236,21 @@ name returned by C<basename($url)>.
 ###############################################
 sub download {
 ###############################################
+    my($url) = @_;
 
     local($Log::Log4perl::caller_depth) += 1;
 
-    _confirm("Downloading $_[0] => ", basename($_[0])) or return 1;
+    INFO "download $url";
 
-    getstore($_[0], basename($_[0])) or 
-        get_logger("")->logcroak("Cannot download $_[0] ($!)");
+    _confirm("Downloading $url => ", basename($url)) or return 1;
+
+    my $rc = getstore($url, basename($_[0]));
+    
+    if($rc != RC_OK) {
+        LOGCROAK("Cannot download $_[0] ($!)");
+    }
+
+    return 1;
 }
 
 =pod
@@ -262,7 +276,7 @@ sub untar {
 ###############################################
     local($Log::Log4perl::caller_depth) += 1;
 
-    get_logger("")->logcroak("untar called without defined tarfile") unless 
+    LOGCROAK("untar called without defined tarfile") unless 
          @_ == 1 and defined $_[0];
 
     _confirm "untar $_[0]" or return 1;
@@ -284,9 +298,9 @@ sub untar {
             # extract as topdir
         $arch->extract();
         rename $topdir, $namedir or 
-            get_logger("")->logcroak("Can't rename $topdir, $namedir");
+            LOGCROAK("Can't rename $topdir, $namedir");
     } else {
-        get_logger("")->logcroak("no topdir") unless defined $topdir;
+        LOGCROAK("no topdir") unless defined $topdir;
         DEBUG "Not-so-nice archive (no topdir), extracting to subdir $topdir";
         $topdir = basename $topdir;
         rmf($topdir);
@@ -315,7 +329,7 @@ sub untar_in {
 
     local($Log::Log4perl::caller_depth) += 1;
 
-    get_logger("")->logcroak("not enough arguments") if
+    LOGCROAK("not enough arguments") if
       ! defined $tar_file or ! defined $dir;
 
     _confirm "Untarring $tar_file in $dir" or return 1;
@@ -330,7 +344,7 @@ sub untar_in {
     require Archive::Tar;
     my $arch = Archive::Tar->new("$tar_file_abs");
     $arch->extract() or 
-        get_logger("")->logcroak("Extract failed: $!");
+        LOGCROAK("Extract failed: ($!)");
     cdback();
 }
 
@@ -373,7 +387,7 @@ sub pick {
     my %files;
 
     if(@_ != 3 or ref($options) ne "ARRAY") {
-        get_logger("")->logcroak("pick called with wrong #/type of args");
+        LOGCROAK("pick called with wrong #/type of args");
     }
     
     {
@@ -385,7 +399,8 @@ sub pick {
             $files{$count} = $_;
         }
     
-        print STDERR "$prompt [$default_int]> ";
+        print STDERR "$prompt [$default_int]> "
+            or die "Couldn't write STDERR: ($!)";
         my $input = <STDIN>;
         chomp($input) if defined $input;
 
@@ -415,10 +430,12 @@ sub ask {
     local($Log::Log4perl::caller_depth) += 1;
 
     if(@_ != 2) {
-        get_logger("")->logcroak("ask() called with wrong # of args");
+        LOGCROAK("ask() called with wrong # of args");
     }
 
-    print STDERR "$prompt [$default]> ";
+    print STDERR "$prompt [$default]> "
+        or die "Couldn't write STDERR: ($!)";
+
     my $value = <STDIN>;
     chomp $value;
 
@@ -443,8 +460,10 @@ sub mkd {
 
     _confirm "mkd @_" or return 1;
 
+    INFO "mkpath @_";
+
     mkpath @_ or 
-        get_logger("")->logcroak("Cannot mkdir @_ ($!)");
+        LOGCROAK("Cannot mkdir @_ ($!)");
 }
 
 =pod
@@ -469,15 +488,21 @@ sub rmf {
         return;
     }
 
+    INFO "rmtree @_";
+
     rmtree $_[0] or 
-        get_logger("")->logcroak("Cannot rmtree $_[0] ($!)");
+        LOGCROAK("Cannot rmtree $_[0] ($!)");
 }
 
 =pod
 
 =item C<cd($dir)>
 
-chdir to the given directory.
+chdir to the given directory. If you don't want to have cd() modify
+the internal directory stack (used for subsequent cdback() calls), 
+set the stack_update parameter to a false value:
+
+    cd($dir, {stack_update => 0});
 
 =cut
 
@@ -488,9 +513,19 @@ sub cd {
     local($Log::Log4perl::caller_depth) += 1;
     INFO "cd $_[0]";
 
-    push @DIR_STACK, getcwd();
+    my $opts = { stack_update => 1 };
+    $opts = $_[1] if ref $_[1] eq "HASH";
+
+    if ($opts->{stack_update}) {
+        my $cwd = getcwd();
+        if(! defined $cwd) {
+            LOGCROAK("Cannot getcwd ($!)");        ;
+        }
+        push @DIR_STACK, $cwd;
+    }
+
     chdir($_[0]) or 
-        get_logger("")->logcroak("Cannot cd $_[0] ($!)");
+        LOGCROAK("Cannot cd $_[0] ($!)");
 }
 
 =pod
@@ -507,11 +542,15 @@ sub cdback {
 
     local($Log::Log4perl::caller_depth) += 1;
 
-    get_logger("")->logcroak("cd stack empty") unless @DIR_STACK;
+    LOGCROAK("cd stack empty") unless @DIR_STACK;
 
     my $old_dir = pop @DIR_STACK;
+
+    LOGCROAK("Directory stack empty")
+        if ! defined $old_dir;
+
     INFO "cdback to $old_dir";
-    cd($old_dir);
+    cd($old_dir, {stack_update => 0});
 }
 
 =pod
@@ -530,8 +569,10 @@ sub make {
 
     _confirm "make @_" or return 1;
 
+    INFO "make @_";
+
     system("make @_") and 
-        get_logger("")->logcroak("Cannot make @_ ($!)");
+        LOGCROAK("Cannot make @_ ($!)");
 }
 
 =pod
@@ -546,7 +587,7 @@ sub check_zlib {
     if($tar_file =~ /\.tar\.gz\b|\.tgz\b/ and
        !Log::Log4perl::Util::module_available("IO::Zlib")) {
 
-        get_logger("")->logcroak("$tar_file: Compressed tarballs can ",
+        LOGCROAK("$tar_file: Compressed tarballs can ",
                "only be processed with IO::Zlib installed.");
     }
 }
@@ -575,8 +616,8 @@ sub archive_sniff {
     my $tar = Archive::Tar->new($name);
 
     my @names = $tar->list_files(["name"]);
-
-    get_logger("")->logcroak("Archive $name is empty") unless @names;
+    
+    LOGCROAK("Archive $name is empty") unless @names;
 
     (my $archdir = $names[0]) =~ s#/.*##;
 
@@ -629,7 +670,7 @@ sub pie {
         my $out = "";
 
         open FILE, "<$file" or 
-            get_logger("")->logcroak("Cannot open $file ($!)");
+            LOGCROAK("Cannot open $file ($!)");
         while(<FILE>) {
             $out .= $coderef->($_);
         }
@@ -669,7 +710,7 @@ sub plough {
         my $out = "";
 
         open FILE, "<$file" or 
-            get_logger("")->logcroak("Cannot open $file ($!)");
+            LOGCROAK("Cannot open $file ($!)");
         while(<FILE>) {
             $coderef->($_);
         }
@@ -703,7 +744,7 @@ sub slurp {
     if($from_file) {
         INFO "Slurping data from $file";
         open FILE, "<$file" or 
-            get_logger("")->logcroak("Cannot open $file ($!)");
+            LOGCROAK("Cannot open $file ($!)");
         $data = <FILE>;
         close FILE;
         DEBUG "Read ", snip($data, $DATA_SNIPPED_LEN), " from $file";
@@ -733,20 +774,56 @@ sub blurt {
 
     local($Log::Log4perl::caller_depth) += 1;
 
+    $append = 0 unless defined $append;
+
     _confirm(($append ? "Appending" : "Writing") . " " .
          length($data) . " bytes to $file") or return 1;
 
     open FILE, ">" . ($append ? ">" : "") . $file 
         or 
-        get_logger("")->logcroak("Cannot open $file for writing ($!)");
-    print FILE $data;
-    close FILE;
+        LOGCROAK("Cannot open $file for writing ($!)");
+    print FILE $data
+        or 
+        LOGCROAK("Cannot write to $file ($!)");        
+    close FILE
+        or 
+        LOGCROAK("Cannot close $file ($!)");        
 
     DEBUG "Wrote ", snip($data, $DATA_SNIPPED_LEN), " to $file";
 }
 
 =pod
 
+=item C<blurt_atomic($data, $file)>
+
+Write the data in $data to a file $file, guaranteeing that the operation
+will either complete fully or not at all. This is accomplished by first
+writing to a temporary file which is then rename()ed to the target file.
+
+Unlike in C<blurt>, there is no C<$append> mode in C<blurt_atomic>.
+
+=cut
+
+###############################################
+sub blurt_atomic {
+###############################################
+    my($data, $file) = @_;
+
+    _confirm("Writing atomically " .
+         length($data) . " bytes to $file") or return 1;
+
+    my($fh, $tmpname) = tempfile(DIR => dirname($file));
+
+    blurt($data, $tmpname);
+
+    rename $tmpname, $file or
+        LOGDIE "Can't rename $tmpname to $file";
+
+    DEBUG "Wrote ", snip($data, $DATA_SNIPPED_LEN), " atomically to $file";
+}
+
+=pod
+
 =item C<($stdout, $stderr, $exit_code) = tap($cmd, @args)>
 
 Run a command $cmd in the shell, and pass it @args as args.
@@ -814,14 +891,14 @@ sub tap {
         $cmd = join ' ', map { qquote($_, ":shell") } @args;
     } else {
             # Default mode: Single quotes
-        $cmd = join ' ', map { quote($_) } @args;
+        $cmd = join ' ', map { quote($_, ":shell") } @args;
     }
        
     $cmd = "$cmd 2>$tmpfile |";
     INFO "tapping $cmd";
 
     open PIPE, $cmd or 
-        get_logger("")->logcroak("open $cmd | failed ($!)");
+        LOGCROAK("open $cmd | failed ($!)");
     my $stdout = join '', <PIPE>;
     close PIPE;
 
@@ -829,6 +906,8 @@ sub tap {
 
     my $stderr = slurp($tmpfile);
 
+    DEBUG "tap $cmd results: rc=$exit_code stderr=[$stderr] stdout=[$stdout]";
+
     return ($stdout, $stderr, $exit_code);
 }
 
@@ -930,7 +1009,37 @@ sub qquote {
 
 =item C<$quoted_string = quote($string, [$metachars])>
 
-Similar to C<qquote()>, just puts a string in single quotes.
+Similar to C<qquote()>, just puts a string in single quotes and
+escapes what needs to be escaped.
+
+Note that shells typically don't support escaped single quotes within
+single quotes, which means that
+
+    $ echo 'foo\'bar'
+    >
+
+is invalid and the shell waits until it finds a closing quote. 
+Instead, there is an evil trick which gives the desired result:
+
+    $ echo 'foo'\''bar'  # foo, single quote, \, 2 x single quote, bar
+    foo'bar
+
+It uses the fact that shells interpret back-to-back strings as one.
+The construct above consists of three back-to-back strings: 
+
+    (1) 'foo'
+    (2) '
+    (3) 'bar'
+
+which all get concatenated to a single 
+
+    foo'bar
+
+If you call C<quote()> with C<$metachars> set to ":shell", it will
+perform that magic behind the scenes:
+
+    print quote("foo'bar");
+      # prints: 'foo'\''bar'
 
 =cut
 
@@ -939,10 +1048,14 @@ sub quote {
 ###############################################
     my($str, $metas) = @_;
 
-    $str =~ s/([\\'])/\\$1/g;
+    if(defined $metas and $metas eq ":shell") {
+        $str =~ s/([\\])/\\$1/g;
+        $str =~ s/(['])/'\\''/g;
+    } else {
+        $str =~ s/([\\'])/\\$1/g;
+    }
 
-    if(defined $metas) {
-        $metas = '' if $metas eq ":shell";
+    if(defined $metas and $metas ne ":shell") {
         $metas =~ s/\]/\\]/g;
         $str =~ s/([$metas])/\\$1/g;
     }
@@ -969,7 +1082,7 @@ sub perm_cp {
 
     _confirm "perm_cp @_" or return 1;
 
-    get_logger("")->logcroak("usage: perm_cp src dst ...") if @_ < 2;
+    LOGCROAK("usage: perm_cp src dst ...") if @_ < 2;
 
     my $perms = perm_get($_[0]);
     perm_set($_[1], $perms);
@@ -994,7 +1107,7 @@ sub perm_get {
 
     my @stats = (stat $filename)[2,4,5] or
         
-        get_logger("")->logcroak("Cannot stat $filename ($!)");
+        LOGCROAK("Cannot stat $filename ($!)");
 
     INFO "perm_get $filename (@stats)";
 
@@ -1022,10 +1135,10 @@ sub perm_set {
 
     chown($perms->[1], $perms->[2], $filename) or 
         
-        get_logger("")->logcroak("Cannot chown $filename ($!)");
+        LOGCROAK("Cannot chown $filename ($!)");
     chmod($perms->[0] & 07777,    $filename) or
         
-        get_logger("")->logcroak("Cannot chmod $filename ($!)");
+        LOGCROAK("Cannot chmod $filename ($!)");
 }
 
 =pod
@@ -1047,10 +1160,10 @@ sub sysrun {
 
     _confirm "sysrun: @cmds" or return 1;
 
-    get_logger("")->logcroak("usage: sysrun cmd ...") if @_ < 1;
+    LOGCROAK("usage: sysrun cmd ...") if @_ < 1;
 
     system(@cmds) and 
-        get_logger("")->logcroak("@cmds failed ($!)");
+        LOGCROAK("@cmds failed ($!)");
 }
 
 =pod
@@ -1128,9 +1241,9 @@ sub sudo_me {
     if($> != 0) {
         DEBUG "Not running as root, calling sudo $0 @$argv";
         my $sudo = bin_find("sudo");
-        get_logger("")->logcroak("Can't find sudo in PATH") unless $sudo;
+        LOGCROAK("Can't find sudo in PATH") unless $sudo;
         exec($sudo, $0, @$argv) or 
-            get_logger("")->logcroak("exec failed!");
+            LOGCROAK("exec failed!");
     }
 }
 
@@ -1178,10 +1291,10 @@ sub fs_read_open {
     local($Log::Log4perl::caller_depth) += 1;
 
     my $find = bin_find("find");
-    get_logger("")->logcroak("Cannot find 'find'") unless defined $find;
+    LOGCROAK("Cannot find 'find'") unless defined $find;
 
     my $cpio = bin_find("cpio");
-    get_logger("")->logcroak("Cannot find 'cpio'") unless defined $cpio;
+    LOGCROAK("Cannot find 'cpio'") unless defined $cpio;
 
     cd $dir;
  
@@ -1189,7 +1302,7 @@ sub fs_read_open {
 
     DEBUG "Reading from $cmd";
     open my $in, "$cmd |" or 
-        get_logger("")->logcroak("Cannot open $cmd");
+        LOGCROAK("Cannot open $cmd");
 
     cdback;
 
@@ -1217,7 +1330,7 @@ sub fs_write_open {
     local($Log::Log4perl::caller_depth) += 1;
 
     my $cpio = bin_find("cpio");
-    get_logger("")->logcroak("Cannot find 'cpio'") unless defined $cpio;
+    LOGCROAK("Cannot find 'cpio'") unless defined $cpio;
 
     mkd $dir unless -d $dir;
 
@@ -1227,7 +1340,7 @@ sub fs_write_open {
 
     DEBUG "Writing to $cmd in dir $dir";
     open my $out, "| $cmd" or 
-        get_logger("")->logcroak("Cannot open $cmd");
+        LOGCROAK("Cannot open $cmd");
 
     cdback;
 
@@ -1254,12 +1367,16 @@ sub pipe_copy {
     my $bytes = 0;
 
     INFO "Opening pipe (bufsize=$bufsize)";
-
-    while(sysread($in, my $buf, $bufsize)) {
+    my $ret;
+    while($ret = sysread($in, my $buf, $bufsize)) {
         $bytes += length $buf;
-        syswrite $out, $buf;
+        if (!defined syswrite $out, $buf) {
+            LOGCROAK("Write to pipe failed: ($!)");
+        }
+    }
+    if (!defined $ret) {
+        LOGCROAK("Read from pipe failed: ($!)");
     }
-
     INFO "Closed pipe (bufsize=$bufsize, transferred=$bytes)";
 }
 
@@ -1320,6 +1437,116 @@ sub printable {
 
 =pod
 
+=item C<password_read($prompt)>
+
+Reads in a password to be typed in by the user in noecho mode.
+A call to password_read("password: ") results in
+
+    password: ***** (stars aren't actually displayed)
+
+This function will switch the terminal back into normal mode
+after the user hits the 'Return' key.
+
+=cut
+
+###########################################
+sub password_read {
+###########################################
+    my($prompt) = @_;
+
+    use Term::ReadKey;
+    ReadMode 'noecho';
+    $| = 1;
+    print "$prompt"
+        or die "Couldn't write STDOUT: ($!)";
+    my $pw = ReadLine 0;
+    chomp $pw;
+    ReadMode 'restore';
+    print "\n"
+        or die "Couldn't write STDOUT: ($!)";
+
+    return $pw;
+}
+
+=pod
+
+=item C<nice_time($time)>
+
+Format the time in a human-readable way, less wasteful than the 
+'scalar localtime' formatting. 
+
+    print nice_time(), "\n";
+      # 2007/04/01 10:51:24
+
+It uses the system time by default, but it can also accept epoch seconds:
+
+    print nice_time(1170000000), "\n";
+      # 2007/01/28 08:00:00
+
+It uses localtime() under the hood, so the outcome of the above will
+depend on your local time zone setting.
+
+=cut
+
+###########################################
+sub nice_time {
+###########################################
+    my($time) = @_;
+
+    $time = time() unless defined $time;
+
+    my ($sec,$min,$hour,$mday,$mon,$year,
+     $wday,$yday,$isdst) = localtime($time);
+
+    return sprintf("%d/%02d/%02d %02d:%02d:%02d",
+     $year+1900, $mon+1, $mday,
+     $hour, $min, $sec);
+}
+
+=item C<def_or($foo, $default)>
+
+Perl-5.9 added the //= construct, which helps assigning values to
+undefined variables. Instead of writing
+
+    if(!defined $foo) {
+        $foo = $default;
+    }
+
+you can just write
+
+    $foo //= $default;
+
+However, this is not available on older perl versions (although there's 
+source filter solutions). Often, people use
+
+    $foo ||= $default;
+
+instead which is wrong if $foo contains a value that evaluates as false.
+So Sysadm::Install, the everything-and-the-kitchen-sink under the CPAN
+modules, provides the function C<def_or()> which can be used like
+
+    def_or($foo, $default); 
+
+to accomplish the same as
+
+    $foo //= $default;
+
+How does it work, how does $foo get a different value, although it's 
+apparently passed in by value? Modifying $_[0] within the subroutine
+is an old Perl trick to do exactly that.
+
+=cut
+
+###########################################
+sub def_or($$) {
+###########################################
+    if(! defined $_[0]) {
+        $_[0] = $_[1];
+    }
+}
+
+=pod
+
 =back
 
 =head1 AUTHOR
@@ -1328,7 +1555,7 @@ Mike Schilli, E<lt>m@perlmeister.comE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2004 by Mike Schilli
+Copyright (C) 2004-2007 by Mike Schilli
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.3 or,
